#include <poll.h>
#include <math.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/timerfd.h>
#ifndef BUILD_FOR_RHEL6
#include <systemd/sd-daemon.h>
#endif

#include "netfilter_platform.h"

#define POLL_MAX_FDS 2
#define MDE_NETFILTER_SOCKET_PATH "/var/run/mde_netfilter.sock"

/* Global structure to maintain netfilter queue handle */
struct net_session *net_sess_handle;

/* initialization for client */
volatile int init_completed = 0;

/* is client active check if not then shutdown */
volatile int should_shutdown = 0;

/* packet timeout 500ms */
#define AUTO_APPROVE_TIMEOUT (500)

/* poll watch fd types */
enum POLL_EVENT_FD_TYPE
{
    POLL_LISTEN_SOCKET_FD = 0,
    POLL_TIMER_FD,
};

/*
 * On RHEL-6 create socket with conventional method. On higher versions of distro versions
 * make use of socket based activation
 */
static int create_named_socket(const char *filename)
{
    int ret;
    int sock = -1;
    struct sockaddr_un addr;

#ifndef BUILD_FOR_RHEL6
    int n = sd_listen_fds(0);
    CHECK_ERROR(n < 0, "Failed to read listening FD from environment");
    CHECK_ERROR(n > 1, "Too many file descriptors received!");

    if (n == 1)
    {
        sock = SD_LISTEN_FDS_START + 0;
        ret = sd_is_socket_unix(sock, SOCK_STREAM, 1, filename, 0);
        CHECK_ERROR(ret < 0, "Failed to verify socket");
    }
    else
    {
#endif

    INFO("Creating socket traditionally");

    /* create the socket */
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    CHECK_ERROR(sock < 0, "Failed to create socket");

    /* fill up socket address strucure */
    addr.sun_family = AF_UNIX;
    strncpy(addr.sun_path, filename, sizeof(addr.sun_path));
    addr.sun_path[sizeof(addr.sun_path) - 1] = '\0';

    /* unlink previous instance/address */
    (void)unlink(MDE_NETFILTER_SOCKET_PATH);

    /* bind to socket address */
    ret = bind(sock, (struct sockaddr *) &addr, sizeof(addr));
    CHECK_ERROR(ret < 0, "Failed to bind socket error: %s", strerror(errno));

    /* increase the listen queue to max 2 */
    listen(sock, 2);

#ifndef BUILD_FOR_RHEL6
    }
#endif  

exit:
    return sock;
}

#ifdef BUILD_FOR_RHEL6
static int daemonize()
{
    int ret = -1;
    pid_t sid = 0;
    pid_t pid = 0;
    
    pid = fork();
    CHECK_ERROR(pid < 0, "Failed to fork new process: %s", strerror(errno));
    
    /* exit the parent process */
    if (pid > 0)
    {
        exit(0);
    }

    /* unmask the file mode */
    umask(0022);

    /* create a new session, to run as session leader */
    sid = setsid();
    CHECK_ERROR(sid < 0, "Failed to set session id: %s", strerror(errno));

    /* change the current working directory to root */
    chdir("/");

    /* close stdin, stdout and stderr */
    close(STDIN_FILENO);
    close(STDOUT_FILENO);
    close(STDERR_FILENO);
    ret = 0;

exit:
    return ret;
}
#endif

static void setup_poll(struct pollfd pfd[], int listenfd, int timerfd)
{
    pfd[POLL_LISTEN_SOCKET_FD].fd = listenfd;
    pfd[POLL_LISTEN_SOCKET_FD].events = POLLIN;
    pfd[POLL_LISTEN_SOCKET_FD].revents = 0;

    pfd[POLL_TIMER_FD].fd = timerfd;
    pfd[POLL_TIMER_FD].events = POLLIN;
    pfd[POLL_LISTEN_SOCKET_FD].revents = 0;
}

static int setup_timer()
{
    int fd = -1;
    int ret = -1;

    struct itimerspec timer_period = 
    {
        .it_interval =
        {
            .tv_sec = 1,
            .tv_nsec = 0
        },
        .it_value =
        {
            .tv_sec = 1,
            .tv_nsec = 0
        }
    };
    fd = timerfd_create(CLOCK_MONOTONIC, TFD_CLOEXEC);
    CHECK_ERROR(fd == -1, "Failed to create timer: %s", strerror(errno));

    ret = timerfd_settime(fd, 0, &timer_period, NULL);
    CHECK_ERROR(ret == -1, "Failed to setup timer: %s", strerror(errno));

exit:
    if (ret == -1 && fd >= 0)
    {
        close(fd);
        fd = -1;
    }
    return fd;
}

uint64_t get_time_in_millis()
{
    uint64_t time_in_millis;

#ifdef BUILD_FOR_RHEL6
    struct timeval current_time = {0};
    gettimeofday(&current_time, NULL);
    time_in_millis = current_time.tv_sec*1000 + lround(current_time.tv_usec/1.0e3);
#else
    struct timespec current_time = {0};
    clock_gettime(CLOCK_MONOTONIC, &current_time);
    time_in_millis = current_time.tv_sec*1000 + lround(current_time.tv_nsec/1.0e6);
#endif

    return time_in_millis;
}

static void ack_timer(int fd)
{
    int ret;
    uint64_t var;

    /* Read timer period event */
    ret = read(fd, &var, sizeof(uint64_t));
    if (ret == -1)
    {
        ERROR("Failed to read from timerfd");
    }
}

/* triggers on every timeout */
static gboolean timeout_check(gpointer key, gpointer value, gpointer user_obj)
{
    uint64_t current_time = *(uint64_t *)user_obj;
    struct outstanding_pkt *packet = (struct outstanding_pkt *)value;

    if ((packet == NULL) || (key == value))
    {
        return FALSE;
    }

    if ((current_time - packet->pkt_time) >= AUTO_APPROVE_TIMEOUT)
    {
        INFO("TIMEOUT! Approving outstanding packet: %d", packet->packetid);
        nfq_set_verdict2(net_sess_handle->queue_ctx.qhandle, packet->packetid, NF_REPEAT, packet->pkt_mark, 0, NULL);

        /* It should return TRUE if the key/value pair should be removed from the GHashTable */
        return TRUE;
    }
    return FALSE;
}

/* destructor for the object in packet hashtable */
void packet_destructor(gpointer object)
{
    struct outstanding_pkt *packet = (struct outstanding_pkt *)object;
    if (packet)
    {
        DEBUG("Cleaning up Packet: %d", packet->packetid);
        free(packet);
    }
}

int main(int argc, char* argv[])
{
    int ret;
    socklen_t len;
    int sock, new_connection;
    client_ctx client = {0};
    struct pollfd poll_fds[POLL_MAX_FDS] = {0};
    struct sockaddr_un remote = {0};
    pthread_t netfilter_event_handler_thread;
    pthread_t client_verdict_recv_thread;

#ifdef BUILD_FOR_RHEL6
    ret = daemonize();
    CHECK_ERROR(ret != 0, "Failed to daemonize the process");
#endif
    INFO("MDE Netfilter process started PID: %d", getpid());

    /* create session object */
    net_sess_handle = (struct net_session *)malloc(sizeof(struct net_session));
    CHECK_ERROR(!net_sess_handle, "Failed to allocate memory for vnet session");
    memset(net_sess_handle, 0, sizeof(struct net_session));

    /* initialize data structure to hold packets */
    pthread_mutex_init(&net_sess_handle->pkt_hashtable_lock, NULL);
    net_sess_handle->pkt_hashtable = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, packet_destructor);

    /* starting service with socket based activation */
    sock = create_named_socket(MDE_NETFILTER_SOCKET_PATH);
    CHECK_ERROR(sock < 0, "Failed to listen on the socket");

    /* setup timer */
    net_sess_handle->timerfd = setup_timer();
    CHECK_ERROR(net_sess_handle->timerfd == -1, "Failed to setup timer");

    /* setup poll */
    setup_poll(poll_fds, sock, net_sess_handle->timerfd);

    while (!ATOMIC_LOAD(&should_shutdown))
    {
        ret = poll(poll_fds, POLL_MAX_FDS, -1);
        if (ret < 0 && errno != EINTR)
        {
            ERROR("Poll interrupted: %s", strerror(errno));
            continue;
        }

        if (poll_fds[POLL_LISTEN_SOCKET_FD].revents & POLLIN)
        {
            /* accept new connection */
            new_connection = accept(sock, (struct sockaddr *)&remote, &len);
            CHECK_ERROR(new_connection < 0, "Failed to accept new connection error: %s", strerror(errno));
            client.queue_num = 0;
            client.fd = new_connection;

            INFO("Client connected: %d.\n", client.fd);

            /* fill up the client information */
            net_sess_handle->clients[0] = client;

            if (!init_completed)
            {
                /* register a callback with netfilter queue library */
                ret = netfilter_platform_init(net_sess_handle);
                CHECK_ERROR(ret < 0, "Error in intialising netfilter queue library handle");

                /* receiver thread - receives packets from netfilter libraries and send to clients */
                ret = pthread_create(&netfilter_event_handler_thread,
                                    NULL,
                                    netfilter_event_handler,
                                    (void *)net_sess_handle);
                CHECK_ERROR(ret != 0, "Could not create netfilter_event_handler thread");

                ret = pthread_create(&client_verdict_recv_thread, NULL, client_verdict_receiver, (void *)net_sess_handle);
                init_completed = 1;
            }
        }

        if (poll_fds[POLL_TIMER_FD].revents & POLLIN)
        {
            uint64_t time_in_millis;
            ack_timer(poll_fds[POLL_TIMER_FD].fd);

            if (!ATOMIC_LOAD(&should_shutdown) && init_completed)
            {                
                time_in_millis = get_time_in_millis();
                pthread_mutex_lock(&net_sess_handle->pkt_hashtable_lock);
                g_hash_table_foreach_remove(net_sess_handle->pkt_hashtable, timeout_check, &time_in_millis);
                pthread_mutex_unlock(&net_sess_handle->pkt_hashtable_lock);
            }
            else
            {
                INFO("Client is not yet connected - Either started manually or reboot sequence");
                goto exit;
            }
        }
    }

exit:
    INFO("Shutting down MDE Netfilter process PID: %d", getpid());
    /* wait for everyone to close gracefully */
    sleep(1);

    /* call netfilter library cleanup function */
    net_cleanup(net_sess_handle);

    /* destroy packet hashtable */
    g_hash_table_destroy(net_sess_handle->pkt_hashtable);
    pthread_mutex_destroy(&net_sess_handle->pkt_hashtable_lock);

    /* free struct net_session */
    if (net_sess_handle) {
        free(net_sess_handle);
        net_sess_handle = NULL;
    }

    return EXIT_SUCCESS;
}
