#include <poll.h>
#include <math.h>
#include <errno.h>
#include <unistd.h>

#include "netfilter_platform.h"

/* Maximum queue length to handle connection burst */
#define VNET_NFQLENGTH 65536

/* Packet size to be retrieved from netfilter queue */
#define VNET_BUFSIZE 1024

/* Mark on packets to be put by our daemon after processing. */
#define NFQ_PKT_MARK  0x1

/* Loop back or local address and mask */
#define LOOPBACK_ADDRESS "127.0.0.1"
#define LOOPBACK_MASK 0xffffffff

/* poll timeout 500ms */
#define POLL_RECV_TIMEOUT (500)

/* Global structure to maintain netfilter queue handle */
extern struct net_session *net_sess_handle;

extern volatile int should_shutdown;

/*
 * Preserve the existing mark of packet and append it with
 * notifier mark
 */
static inline uint32_t update_packet_mark(struct nfq_data *nfad)
{
   return nfq_get_nfmark(nfad) | NFQ_PKT_MARK;
}

/* Log ipv 4/6 address */
void log_ip_address(const struct sockaddr_storage *addr)
{
    char buffer[INET6_ADDRSTRLEN] = {0};

    if (!addr) {
        return;
    }

    switch (addr->ss_family)
    {
        case AF_INET:
        {
            const struct sockaddr_in* ip4_addr = (struct sockaddr_in*)addr;
            inet_ntop(AF_INET, (const void *)&ip4_addr->sin_addr, buffer, INET_ADDRSTRLEN);
            DEBUG("addr: %s, port: %u\n", buffer, ip4_addr->sin_port);
        }
        break;

        case AF_INET6:
        {
            const struct sockaddr_in6* ip6_addr = (struct sockaddr_in6*)addr;
            inet_ntop(AF_INET6, (const void *)&ip6_addr->sin6_addr, buffer, INET6_ADDRSTRLEN);
            DEBUG("addr: %s, port: %u\n", buffer, ip6_addr->sin6_port);
        }
        break;

        default:
        {
            ERROR("Not supported IP address %d\n", addr->ss_family);
        }
    }
}

/*
 * get inbound/outbound event type based on network hook embedded in packet
 * header
 */
enum conn_event_type get_event_type(struct nfq_data *nfa)
{
   uint32_t index;
   enum nf_inet_hooks hook;
   struct nfqnl_msg_packet_hdr *ph = NULL;
   enum conn_event_type ret = MAX_EVENT;

   if (!nfa) {
      goto exit;
   }

   /* Get the nfq message packet header */
   ph = nfq_get_msg_packet_hdr(nfa);
   if (NULL == ph) {
      goto exit;
   }
   hook = ph->hook;

   /*
    * The index of the device the queued packet was received via.
    * If the returned index is 0, the packet was locally generated or the input
    * interface is not known (i.e. POSTROUTING)
    */
   index = nfq_get_indev(nfa);
   if (index && (NF_INET_LOCAL_IN == hook)) {
      ret = INBOUND_PRECONNECT;
      goto exit;
   }

   /*
    * The index of the device the queued packet will be sent out.  If the
    * returned index is 0, the packet is destined for localhost or the output
    * interface is not yet known (ie. PREROUTING?).
    */
   index = nfq_get_outdev(nfa);
   if (index && (NF_INET_LOCAL_OUT == hook)) {
      ret = OUTBOUND_PRECONNECT;
      goto exit;
   }

   DEBUG("Packet received for invalid hook %d", (int)hook);

exit:
   return ret;
}

/*
 * send packet to client for vetting
 */
int send_to_client(uint32_t packet_mark, 
                  const struct conn_identity_data *conn_data,
                  struct net_session *sess)
{
   ssize_t ret = 0;
   client_ctx client = {0};
   struct outstanding_pkt *packet = NULL;

   /* sanity check */
   CHECK_ERROR((!conn_data || !sess), "MDE Netfilter session is not yet initialized");

   /* allocate memory for packet */
   packet = (struct outstanding_pkt *)calloc(1, sizeof(struct outstanding_pkt));
   CHECK_ERROR(!conn_data, "Failed to allocate memory for a packet");

   packet->packetid = conn_data->event_id;
   packet->pkt_mark = packet_mark;
   packet->pkt_time = get_time_in_millis();

   /* insert the packet in hashtablr for bookkeeping */
   pthread_mutex_lock(&sess->pkt_hashtable_lock);
   g_hash_table_insert(sess->pkt_hashtable, GUINT_TO_POINTER(conn_data->event_id), (gpointer)packet);
   pthread_mutex_unlock(&sess->pkt_hashtable_lock);

   /* get client information */
   client = sess->clients[0];
   DEBUG("Send Packet %d Mark: %u Protocol: %u to client", packet->packetid, packet->pkt_mark, conn_data->protocol);
   ret = send(client.fd, conn_data, sizeof(*conn_data), 0);
   CHECK_ERROR(ret < 0, "Failed to provide packt to client for vetting error: %s", strerror(errno));
   
exit:
   return ret;
}

/*
 * Verdict receiver thread - It waits on socket. 
 * Recieves and send the same to netfilter library
 */ 
void* client_verdict_receiver(void *arg)
{
   int ret = -1;
   int sockfd = -1;
   gpointer gkey = NULL;
   gpointer gvalue = NULL;
   gboolean found = FALSE;
   packet_verdict cli_verdict = {0};
   struct outstanding_pkt *pkt_in_hashtable = NULL;
   struct net_session *sess = (struct net_session *)arg;
   CHECK_ERROR(!sess, "MDE Netfilter session is not yet initialized");

   sockfd = sess->clients[0].fd;
   while(!ATOMIC_LOAD(&should_shutdown))
   {
      ret = recv(sockfd, (void *)&cli_verdict, sizeof(packet_verdict), 0);
      CHECK_ERROR(ret == 0, "Client orderly shutdown: %s", strerror(errno));

      if (cli_verdict.packetid == 0 || (cli_verdict.verdict != NF_REPEAT && cli_verdict.verdict != NF_DROP))
      {
         ERROR("Client didn't send proper verdict: %d for the packet: %d", cli_verdict.packetid, cli_verdict.verdict);
         continue;
      }

      pthread_mutex_lock(&sess->pkt_hashtable_lock);
      found = g_hash_table_lookup_extended(sess->pkt_hashtable, GUINT_TO_POINTER(cli_verdict.packetid), &gkey, &gvalue);
      pthread_mutex_unlock(&sess->pkt_hashtable_lock);

      if (found)
      {
         pkt_in_hashtable = (struct outstanding_pkt *)gvalue;

        /*
         * provide NF_REPEAT verdict (and not NF_ACCEPT) with mask bit set
         * so that the packet iterate through rest of the rules in the ipchain
         */
         DEBUG("PacketID id: %u Verdict: %d packet mark: %u", cli_verdict.packetid, cli_verdict.verdict, pkt_in_hashtable->pkt_mark);
         nfq_set_verdict2(sess->queue_ctx.qhandle, cli_verdict.packetid, cli_verdict.verdict, pkt_in_hashtable->pkt_mark, 0, NULL);

         /* check packet entry in packet hashtable */
         pthread_mutex_lock(&sess->pkt_hashtable_lock);
         g_hash_table_remove(sess->pkt_hashtable, GUINT_TO_POINTER(cli_verdict.packetid));
         pthread_mutex_unlock(&sess->pkt_hashtable_lock);
      }
   }

exit:
   if (!ret)
   {
      ATOMIC_STORE(&should_shutdown, 1);
   }
   INFO("Exiting client recevier thread: %d", should_shutdown);
   return NULL;
}

/*
 * Receive the network event notification from conntrack and NFQUEUE and
 * process it.
 */
void* netfilter_event_handler(void *arg)
{
   int ret = -1;
   ssize_t bread = 0;
   struct pollfd poll_fd = {0};
   char buf[VNET_BUFSIZE] __attribute__ ((aligned));
   struct net_session *sess = (struct net_session *)arg;
 
   if (!sess) {
      goto exit;
   }

   INFO("Started netfilter event handling thread");
   
   /* setup poll */
   poll_fd.events = POLLIN;
   poll_fd.fd = sess->queue_ctx.qfd;
   poll_fd.revents = 0;

   while (!ATOMIC_LOAD(&should_shutdown)) {
      ret = poll(&poll_fd, 1, POLL_RECV_TIMEOUT);
      if (ret < 0 && errno != EINTR)  // interrupted, continue
      {
         ERROR("Poll interrupted: %s", strerror(errno));
         continue;
      }

      if (ret == 0) // timeout
      {
         continue;
      }
      else if (poll_fd.revents & POLLIN) // we have data to read
      {
         /* Read the packet from the netlink socket which is in NFQUEUE */
         bread = recv(sess->queue_ctx.qfd, buf, sizeof(buf), 0);
         if (bread > 0) {
            pthread_mutex_lock(&sess->queue_lock);
            nfq_handle_packet(sess->queue_ctx.handle, buf, bread);
            pthread_mutex_unlock(&sess->queue_lock);
         }

         /*
         * To avoid ENOBUFS error , queue max length is increased with
         * nfq_set_queue_maxlen as per the netfilter documentation
         */
         if ((bread < 0) && (ENOBUFS == errno)) {
            ERROR("Dropping packets due to insufficient memory!\n");
            continue;
         }
      }
   }

exit:
   INFO("Netfilter receiver thread exiting");
   return NULL;
}

/*
 * A callback which gets invoked in the context of receiver thread from
 * nfq_handle_packet function
 */
static int net_queue_callback(struct nfq_q_handle *qh,
                       struct nfgenmsg *nfmsg,
                       struct nfq_data *nfa,
                       void *arg)
{
   int ret = -1;
   int verdict = NF_REPEAT;
   uint32_t packet_mark;
   unsigned char *data = NULL;
   struct nfqnl_msg_packet_hdr *ph = NULL;
   struct iphdr *ipinfo = NULL;
   struct ip6_hdr *ip6info = NULL;
   struct tcphdr *tcp_info = NULL;
   struct udphdr *udp_info = NULL;
   struct conn_identity_data *conn_data = NULL;
   struct net_session *sess = (struct net_session*)arg;

   sess = net_sess_handle;
   if (!nfa|| !sess) {
      goto exit;
   }

   conn_data = (struct conn_identity_data *)calloc(1, sizeof(struct conn_identity_data));
   CHECK_ERROR(!conn_data, "Memory allocation failed for msg data");


   ph = nfq_get_msg_packet_hdr(nfa);
   if (ph) {
      /* event_id is identified with packet_id */
      conn_data->event_id = ntohl(ph->packet_id);
      DEBUG("hw_protocol=0x%04x, hook=%u, id=%u", ntohs(ph->hw_protocol),
            ph->hook, conn_data->event_id);
   }

   ret = nfq_get_payload(nfa, &data);
   CHECK_ERROR((-1 == ret || ret > VNET_BUFSIZE), "Invalid packet length: %d, packet id: %u", ret, conn_data->event_id);

   /* Get the iphdr from the payload */
   ipinfo = (struct iphdr *)data;

   if (IPVERSION == ipinfo->version) {
      conn_data->src.ss_family = AF_INET;
      conn_data->dst.ss_family = AF_INET;
      struct sockaddr_in *src = (struct sockaddr_in *)&conn_data->src;
      struct sockaddr_in *dst = (struct sockaddr_in *)&conn_data->dst;
      src->sin_addr.s_addr = ipinfo->saddr;
      dst->sin_addr.s_addr = ipinfo->daddr;

      if (IPPROTO_TCP == ipinfo->protocol) {
         /* Get the tcphdr from the payload */
         tcp_info = (struct tcphdr *)(data + sizeof(*ipinfo));
         src->sin_port = ntohs(tcp_info->source);
         dst->sin_port = ntohs(tcp_info->dest);

      } else if (IPPROTO_UDP == ipinfo->protocol) {
         /* Get the tcphdr from the payload */
         udp_info = (struct udphdr *)(data + sizeof(*ipinfo));
         src->sin_port = ntohs(udp_info->source);
         dst->sin_port = ntohs(udp_info->dest);
      } else {
         WARN("Non tcp/ip traffic: %d, id=%u", ipinfo->protocol, conn_data->event_id);
         goto exit;
      }
      conn_data->protocol = ipinfo->protocol;
      log_ip_address(&conn_data->src);
      log_ip_address(&conn_data->dst);
   } else {
      ip6info = (struct ip6_hdr *)data;
      conn_data->src.ss_family = AF_INET6;
      conn_data->dst.ss_family = AF_INET6;
      struct sockaddr_in6 *src6 = (struct sockaddr_in6 *)&(conn_data->src);
      struct sockaddr_in6 *dst6 = (struct sockaddr_in6 *)&(conn_data->dst);

      if (IPPROTO_TCP == ip6info->ip6_nxt) {
         tcp_info = (struct tcphdr *)(data + sizeof(*ip6info));
         src6->sin6_port = ntohs(tcp_info->source);
         dst6->sin6_port = ntohs(tcp_info->dest);

      } else if (IPPROTO_UDP == ip6info->ip6_nxt) {
         udp_info = (struct udphdr *)(data + sizeof(*ip6info));
         src6->sin6_port = ntohs(udp_info->source);
         dst6->sin6_port = ntohs(udp_info->dest);
      } else {
         WARN("Non tcp/ipv6 traffic: %d, id=%u", ip6info->ip6_nxt, conn_data->event_id);
         goto exit;
      }
      memcpy(src6->sin6_addr.s6_addr,
             ip6info->ip6_src.s6_addr,
             16);
      memcpy(dst6->sin6_addr.s6_addr,
             ip6info->ip6_dst.s6_addr,
             16);

      conn_data->protocol = ip6info->ip6_nxt;
      log_ip_address(&conn_data->src);
      log_ip_address(&conn_data->dst);
   }

   conn_data->event_type = get_event_type(nfa);

   /* Mark the packet with our mark */
   packet_mark = update_packet_mark(nfa);
   ret = -1;

   /* Don't add packet depicting invalid tcp state to the thread pool queue */
   if (tcp_info) {
      if ((!tcp_info->syn) || (tcp_info->ack)) {
         goto exit;
      }
   }

   /* Don't add packet received on invalid hook to the thread pool queue */
   CHECK_ERROR(0 == conn_data->event_type, "Invalid event type, ignoring packet: id=%u", conn_data->event_id);

   /* Deliver packet to client */
   ret = send_to_client(packet_mark, conn_data, sess);

exit:
   /* If there are no clients connected ret = 0 */
   if (ret < 0 && sess) {
      /* default allow the packet in case of fail to process */
      INFO("ALLOW - PacketID id: %d", conn_data->event_id);
      nfq_set_verdict2(sess->queue_ctx.qhandle,
                      conn_data->event_id,
                      verdict,
                      packet_mark,
                      0,
                      NULL);
   }

   if (conn_data) {
      free(conn_data);
   }
   return ret;
}

/* 
 * Register and setup callback with netfilter library
 */
int netfilter_platform_init(struct net_session *sess)
{
   int ret = -1;
   struct nfq_q_handle *qh = NULL;

   /* sanity check*/
   CHECK_ERROR(!sess, "Session is not yet initialized");

   sess->queue_ctx.handle = nfq_open();
   CHECK_ERROR(!sess->queue_ctx.handle, "Error %s during nfq_open()", strerror(errno));

   ret = nfq_unbind_pf(sess->queue_ctx.handle, AF_INET);
   CHECK_ERROR(ret < 0, "Error %s during nfq_unbind_pf() with AF_INET", strerror(errno));

   ret = nfq_bind_pf(sess->queue_ctx.handle, AF_INET);
   CHECK_ERROR(ret < 0, "Error %s  during nfq_bind_pf() with AF_INET", strerror(errno));

   ret = nfq_unbind_pf(sess->queue_ctx.handle, AF_INET6);
   CHECK_ERROR(ret < 0, "Error %s during nfq_unbind_pf() with AF_INET6", strerror(errno));

   ret = nfq_bind_pf(sess->queue_ctx.handle, AF_INET6);
   CHECK_ERROR(ret < 0, "Error %s during nfq_bind_pf() with AF_INET6", strerror(errno));

   ret = -1;
   qh = nfq_create_queue(sess->queue_ctx.handle,
                         0,
                         &net_queue_callback,
                         NULL);
   CHECK_ERROR(!qh, "Error %s during nfq_create_queue()", strerror(errno));

   pthread_mutex_init(&sess->queue_lock, NULL);
   ret = nfq_set_mode(qh, NFQNL_COPY_PACKET, VNET_BUFSIZE);
   CHECK_ERROR(ret < 0, "Error %s during nfq_set_mode", strerror(errno));

   /* Set kernel queue maximum length to handle connection burst */
   ret = nfq_set_queue_maxlen(qh, VNET_NFQLENGTH);
   if (ret < 0) {
      ERROR("Error %s during nfq_set_queue_maxlen()", strerror(errno));
      ret = 0;
      /* don't fail, continue */
   }

   sess->queue_ctx.qhandle = qh;
   sess->queue_ctx.qfd = nfq_fd(sess->queue_ctx.handle);
   INFO("Initialize the netfilter queue with FD: %d",  sess->queue_ctx.qfd);

exit:

   /*
    * cleanup would be done by caller thread
    * context after the completion of thread.
    */
   return ret;
}

/*
 * Unregister with netfilter queue library handle and lock.
 */
void net_cleanup(struct net_session *sess)
{
   /* sanity check*/
   CHECK_ERROR(!sess, "Session is not yet initialized");

   if (sess->queue_ctx.qhandle) {
      nfq_destroy_queue(sess->queue_ctx.qhandle);
      pthread_mutex_destroy(&sess->queue_lock);
   }

   /* Close netfilter queue library handle */
   if (sess->queue_ctx.handle) {
      nfq_close(sess->queue_ctx.handle);
   }

exit:
   return;
}
